;;; -*- Scheme -*-
;;
;;                      Copyright (c) 1998 SQUARE USA, Inc.
;;                             All Rights Reserved.
;;
;;  SQUARE USA Inc. hereby grants permission to use, copy, modify, and
;;  distribute this software and its documentation for any purpose, provided
;;  that existing copyright notices are retained verbatim in all copies.
;;
;;  IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
;;  FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
;;  ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
;;  DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
;;  POSSIBILITY OF SUCH DAMAGE.
;;
;;  THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
;;  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
;;  FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
;;  IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
;;  NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
;;  MODIFICATIONS.
;;  
;;  $Id: html-classes.scm,v 1.1 1999/11/10 11:08:20 shiro Exp $
;;  

;;;
;;; HTML generation
;;;
;;; Kaveh Kardan -- 22 December 1997 -- repackaged for PDBrev2
;;; Modified by Shiro Kawai
;;;

(require "stklos")
(provide "html-classes")

(define *default-http-server* "")

;;================================================================
;; Formatting
;;   safe-html-string STR
;;   safe-url-string  STR

;;; safe-html-string ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (safe-html-string str)
  (let ((result (list)))
    (for-each (lambda (char)
		(cond ((eq? char #\<) (set! result
					    (append '(#\; #\t #\l #\&)
						    result)))
		      ((eq? char #\>) (set! result
					    (append '(#\; #\t #\g #\&)
						    result)))
		      ((eq? char #\&) (set! result
					    (append '(#\; #\p #\m #\a #\&)
						    result)))
		      ((eq? char #\") (set! result
					    (append '(#\; #\t #\o #\u #\q #\&)
						    result)))
		      (#t (set! result (cons char result)))))
	      (string->list str))
    (list->string (reverse result))))

;;; see http://www.w3.org/Protocols/rfc2068/rfc2068 for details
(define (safe-url-string str)
  (let ((result (list)))
    (for-each (lambda (char)
		(cond 
		 ((eq? char #\;) (set! result (append '(#\b #\3 #\%) result)))
		 ((eq? char #\/) (set! result (append '(#\f #\2 #\%) result)))
		 ((eq? char #\?) (set! result (append '(#\f #\3 #\%) result)))
		 ((eq? char #\:) (set! result (append '(#\a #\3 #\%) result)))
		 ((eq? char #\@) (set! result (append '(#\0 #\4 #\%) result)))
		 ((eq? char #\&) (set! result (append '(#\6 #\2 #\%) result)))
		 ((eq? char #\=) (set! result (append '(#\d #\3 #\%) result)))
		 ((eq? char #\+) (set! result (append '(#\b #\2 #\%) result)))
		 ((eq? char #\") (set! result (append '(#\2 #\2 #\%) result)))
		 ((eq? char #\#) (set! result (append '(#\3 #\2 #\%) result)))
		 ((eq? char #\%) (set! result (append '(#\5 #\2 #\%) result)))
		 ((eq? char #\<) (set! result (append '(#\c #\3 #\%) result)))
		 ((eq? char #\>) (set! result (append '(#\e #\3 #\%) result)))
		 ((eq? char #\space) (set! result
					   (append '(#\0 #\2 #\%) result)))
		 (#t (set! result (cons char result)))))
	      (string->list str))
    (list->string (reverse result))))

;;================================================================
;; Mixin classes for tag options
;;   class <html-bgcolor-mixin>
;;   class <html-size-mixin>
;;   class <html-alignment-mixin>
;;   class <html-table-element-mixin>
;;   method write-html-tag-options

;; default-method
(define-method write-html-tag-options ((self <top>))
  #f)

;;; <html-bgcolor-mixin> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-bgcolor-mixin> ()
  ((background-color :accessor background-color
                     :init-keyword :background-color
                     :initform #f))
  )

(define-method write-html-tag-options ((self <html-bgcolor-mixin>))
  (if (background-color self)
      (format #t " bgcolor=~s" (background-color self)))
  (next-method))

;;; <html-size-mixin> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-size-mixin> ()
  ((width  :accessor width  :initform #f :init-keyword :width)
   (height :accessor height :initform #f :init-keyword :height))
  )

(define-method write-html-tag-options ((self <html-size-mixin>))
  (if (width self)  (format #t " width=~s" (width self)))
  (if (height self) (format #t " height=~s" (height self)))
  (next-method))

;;; <html-alignment-mixin>;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-alignment-mixin> ()
  ((alignment            :accessor alignment :initform #f
                         :init-keyword :alignment)
   (horizontal-alignment :accessor horizontal-alignment :initform #f
                         :init-keyword :horizontal-alignment)
   (vertical-alignment   :accessor vertical-alignment :initform #f
                         :init-keyword :vertical-alignment))
  )

(define-method write-html-tag-options ((self <html-alignment-mixin>))
  (if (alignment self)
      (format #t " align=~s" (alignment self)))
  (if (horizontal-alignment self)
      (format #t " halign=~s" (horizontal-alignment self)))
  (if (vertical-alignment self)
      (format #t " valign=~s" (vertical-alignment self)))
  (next-method))

;;; <html-table-element-mixin> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-table-element-mixin> ()
  ((rowspan   :accessor rowspan :initform #f :init-keyword :rowspan)
   (colspan   :accessor colspan :initform #f :init-keyword :colspan)))

(define-method write-html-tag-options ((self <html-table-element-mixin>))
  (if (rowspan self)
      (format #t " rowspan=~s" (rowspan self)))
  (if (colspan self)
      (format #t " colspan=~s" (colspan self)))
  (next-method))

;;; <html-border-mixin> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-border-mixin> ()
  ((border    :accessor border :initform #f :init-keyword :border)))

(define-method write-html-tag-options ((self <html-border-mixin>))
  (if (border self)
      (format #t " border=~s" (border self)))
  (next-method))

;;; <html-target-frame-mixin> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; For a historical reason, a slot for TARGET attribute is called
;;; TARGET-FRAME.

(define-class <html-target-frame-mixin> ()
  ((target-frame :accessor target-frame :initform #f
                 :init-keyword :target-frame)))

(define-method write-html-tag-options ((self <html-target-frame-mixin>))
  (if (target-frame self)
      (format #t " target=~s" (target-frame self)))
  (next-method))

;;================================================================
;; HTML classes
;;   <html-container>
;;     +-- <html-page> (w/ bgcolor)
;;     +-- <html-framed-page>
;;     +-- <html-anchor>
;;     +-- <html-form>
;;   <html-line-break>
;;   <html-java-script>
;;   <html-image> (w/ alignment, size, border)
;;   <html-paragraph-break>
;;   <html-paragraph> (w/ alignment)
;;   <html-text>
;;   <html-header> (w/ alignment)

;; Fallback method of write-html.

(define-method write-html ((obj <top>))
  (display obj))

;;; <html-image> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-image> (<html-alignment-mixin>
                            <html-size-mixin>
                            <html-border-mixin>)
  ((source      :accessor source      :init-keyword :source      :initform "")
   (alternative :accessor alternative :init-keyword :alternative :initform "")
   (hspace      :accessor hspace      :init-keyword :hspace      :initform #f)
   (vspace      :accessor vspace      :init-keyword :vspace :initform #f)))

(define-method write-html-tag-options ((image <html-image>))
  (if (source image) (format #t " src=~s" (source image)))
  (if (alternative image) (format #t " alt=~s" (alternative image)))
  (if (hspace image) (format #t " hspace=~s" (hspace image)))
  (if (vspace image) (format #t " vspace=~s" (vspace image)))
  (next-method))

(define-method write-html ((image <html-image>))
  (format #t "<img")
  (write-html-tag-options image)
  (format #t ">\n"))

(define-macro (html-image . args)
  `(make <html-image> ,@args))

;;; <html-java-script> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-java-script> ()
  ((script-name :accessor script-name :init-keyword :script-name)
   (script-body :accessor script-body :init-keyword :script-body)))
  
(define-method write-html ((java-script <html-java-script>))
  (format #t "<script language=\"JavaScript\">\n")
  (format #t "~a~%" (script-body java-script))
  (format #t "</script>\n"))

(define-macro (html-java-script . args)
  `(make <html-java-script> ,@args))

;;; <html-container> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  Abstract base class for html block tags which have contents.

(define-class <html-container> ()
  ((contents :initform '() :accessor contents :init-keyword :contents)))

(define-method write-html ((container <html-container>))
  (let ((c (contents container)))
    (if (list? c) (map write-html c) (write-html c))))

(define-macro (html-container . args)
  `(make <html-container> ,@args))

;;; <html-document>;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  Individual html documents, i.e. a document surrounded by <html>...
;;;  </html>.   Although it is not checked, CONTENTS field must have either
;;;    * List of one elements, <html-body>.
;;;    * List of one <html-frameset> and (optionally) <html-noframes>.
;;;  The composite class <html-page> is provided for generating simple
;;;  html page.
;;;
;;;  META-INFO and HTTP-EQUIV-INFO are used to generate META tags.
;;;  Both should be a list of (NAME VALUE).   META-INFO generates
;;;  simple <META NAME=...> tag, while HTTP-EQUIV-INFO generates
;;;  <META HTTP-EQUIV=...> tag.
;;;  The most common and useful way to use META HTTP-EQUIV is specifying
;;;  document character set, for example:
;;;    <meta http-equiv="Content-Type" content="text/html; charset=EUC-JP">
;;;
;;;  BASE-HREF and BASE-TARGET are used for <BASE HREF=...> and
;;;  <BASE TARGET=...> tags respectively.

(define-class <html-document> (<html-container>)
  ((page-title         :accessor page-title
                       :init-keyword :page-title
                       :initform #f)
   (meta-info          :accessor meta-info
                       :init-keyword :meta-info
                       :initform '())
   (http-equiv-info    :accessor http-equiv-info
                       :init-keyword :http-equiv-info
                       :initform '())
   (base-href          :accessor base-href
                       :init-keyword :base-href
                       :initform #f)
   (base-target        :accessor base-target
                       :init-keyword :base-target
                       :initform #f)
   ))

(define-method write-html ((self <html-document>))
  (format #t "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML//EN\">\n")
  (format #t "<html>\n")
  (format #t "<!-- This page was automatically generated by PWEB. -->\n")
  (format #t "<head>\n")
  (if (page-title self)
      (format #t "<title>~a</title>\n" (page-title self))
      (format #t "<title>Untitled</title>\n"))
  (map (lambda (n&v)
         (format #t "<meta name=~s content=~s>\n" (car n&v) (cadr n&v)))
       (meta-info self))
  (map (lambda (n&v)
         (format #t "<meta http-equiv=~s content=~s>\n" (car n&v) (cadr n&v)))
       (http-equiv-info self))
  (if (base-href self)
      (format #t "<base href=~s>\n" (base-href self)))
  (if (base-target self)
      (format #t "<base target=~s>\n" (base-target self)))
  (format #t "</head>\n")
  (next-method)
  (format #t "</html>\n"))

(define-macro (html-document . args)
  `(make <html-document> ,@args))

;;; <html-body> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-body> (<html-container> <html-bgcolor-mixin>)
  ((text-color         :accessor text-color
                       :initform #f
                       :init-keyword :text-color)
   (link-color         :accessor link-color
                       :initform #f
                       :init-keyword :link-color)
   (active-link-color  :accessor active-link-color
                       :initform #f
                       :init-keyword :active-link-color)
   (visited-link-color :accessor visited-link-color
                       :initform #f
                       :init-keyword :visited-link-color)
   (background-image   :accessor background-image
                       :initform #f
                       :init-keyword :background-image)
   ;; Following slots are not the attributes of BODY tag, but
   ;; for generating common elements in the content of BODY.
   (body-header        :accessor body-header
                       :init-keyword :body-header
                       :initform #f)
   (author-name        :accessor author-name
                       :init-keyword :author-name
                       :initform #f)
   (author-email       :accessor author-email
                       :init-keyword :author-email
                       :initform #f)
   (signature-string   :initform "This page is maintained by"
		       :accessor signature-string
		       :init-keyword :signature-string)
   ))

(define-method write-html-tag-options ((body <html-body>))
  (if (text-color body)
      (format #t " text=~s" (text-color body)))
  (if (link-color body)
      (format #t " link=~s" (link-color body)))
  (if (active-link-color body)
      (format #t " alink=~s" (active-link-color body)))
  (if (visited-link-color body)
      (format #t " vlink=~s" (visited-link-color body)))
  (if (background-image body)
      (format #t " background=~s" (background-image body)))
  (next-method))

(define-method write-html ((body <html-body>))
  (format #t "<body")
  (write-html-tag-options body)
  (format #t ">\n")
  (if (body-header body) (write-html (body-header body)))
  (next-method)
  (if (author-name body)
      (format #t "<hr>\n<address>~a <a href=\"mailto:~a\">~a</a>.</address>\n"
                (safe-html-string (signature-string body))
                (safe-html-string (author-email body))
                (safe-html-string (author-name body))))
  (format #t "</body>\n"))

(define-macro (html-body . args)
  `(make <html-body> ,@args))

;;; <html-page> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  Object for simple generation of html page.   This is upper-compatible
;;;  with previous versions of schtml. 
;;;  Using composite-metaclass, it looks as if it is an object which has
;;;  features of both <html-document> and <html-body>.
;;;
;;;  If the value of BODY-HEADER is #f, the <H1> header is generated from
;;;  PAGE-TITLE value.

(define-class <html-page> ()
  ((document-object    :initform (make <html-document>))
   (body-object        :initform (make <html-body>))
   (page-title         :accessor page-title
                       :init-keyword :page-title
                       :allocation :propagated
                       :propagate-to (document-object))
   (meta-info          :accessor meta-info
                       :init-keyword :meta-info
                       :allocation :propagated
                       :propagate-to (document-object))
   (http-equiv-info    :accessor http-equiv-info
                       :init-keyword :http-equiv-info
                       :allocation :propagated
                       :propagate-to (document-object))
   (base-href          :accessor base-href
                       :init-keyword :base-href
                       :allocation :propagated
                       :propagate-to (document-object))
   (base-target        :accessor base-target
                       :init-keyword :base-target
                       :allocation :propagated
                       :propagate-to (document-object))
   (contents           :accessor contents
                       :init-keyword :contents
                       :allocation :propagated
                       :propagate-to (body-object))
   (background-color   :accessor background-color
                       :init-keyword :background-color
                       :allocation :propagated
                       :propagate-to (body-object))
   (text-color         :accessor text-color
                       :init-keyword :text-color
                       :allocation :propagated
                       :propagate-to (body-object))
   (link-color         :accessor link-color
                       :init-keyword :link-color
                       :allocation :propagated
                       :propagate-to (body-object))
   (active-link-color  :accessor active-link-color
                       :init-keyword :active-link-color
                       :allocation :propagated
                       :propagate-to (body-object))
   (visited-link-color :accessor visited-link-color
                       :init-keyword :visited-link-color
                       :allocation :propagated
                       :propagate-to (body-object))
   (background-image   :accessor background-image
                       :init-keyword :background-image
                       :allocation :propagated
                       :propagate-to (body-object))
   (body-header        :accessor body-header
                       :init-keyword :body-header
                       :allocation :propagated
                       :propagate-to (body-object))
   (author-name        :accessor author-name
                       :init-keyword :author-name
                       :allocation :propagated
                       :propagate-to (body-object))
   (author-email       :accessor author-email
                       :init-keyword :author-email
                       :allocation :propagated
                       :propagate-to (body-object))
   (signature-string   :accessor signature-string
		       :init-keyword :signature-string
                       :allocation :propagated
                       :propagate-to (body-object))
  )
  :metaclass <Composite-metaclass>)

(define-method initialize ((page <html-page>) initargs)
  (next-method)
  (slot-set! (slot-ref page 'document-object)
             'contents
             (list (slot-ref page 'body-object)))
  )

(define-method write-html ((page <html-page>))
  (if (and (not (body-header page))
           (page-title page))
      (set! (body-header page)
            (make <html-header> :string (page-title page))))
  (write-html (slot-ref page 'document-object)))

(define-macro (html-page . args)
  `(make <html-page> ,@args))

;;; <html-anchor> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Unfortunately, slot name "target" is used to refer to the referring
;;; URL (HREF attribute) in previous versions of schtml.  To retain
;;; backward compatibility, we need to give the name "target-frame" to
;;; the TARGET attribute.   The slot "target" remains as a virtual slot,
;;; but programmers are strongly encouraged to use new slot name "href"
;;; instead.

(define-class <html-anchor> (<html-container> <html-target-frame-mixin>)
  ((href   :accessor href   :init-keyword :href :initform #f)
   (name   :accessor name   :init-keyword :name :initform #f)
   (target :accessor target :init-keyword :target
           :allocation :virtual
           :slot-ref (lambda (o) (slot-ref o 'href))
           :slot-set! (lambda (o v) (slot-set! o 'href v))))
  )

(define-method write-html-tag-options ((self <html-anchor>))
  (if (href self) (format #t " href=~s" (href self)))
  (if (name self) (format #t " name=~s" (name self)))
  (if (target-frame self) (format #t " target=~s" (target-frame self))))

(define-method write-html ((self <html-anchor>))
  (format #t "<a")
  (write-html-tag-options self)
  (format #t ">")
  (next-method)
  (format #t "</a>"))

(define-macro (html-anchor . args)
  `(make <html-anchor> ,@args))

;;; <html-ruler> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-ruler> (<html-alignment-mixin>
                            <html-size-mixin>)
  ((noshade?  :initform #f :accessor noshade? :init-keyword :noshade?)))

(define-method write-html-tag-options ((self <html-ruler>))
  (if (noshade? self) (format #t " noshade"))
  (next-method))

(define-method write-html ((self <html-ruler>))
  (format #t "<hr")
  (write-html-tag-options self)
  (format #t ">"))

(define-macro (html-ruler . args)
  `(make <html-ruler> ,@args))

;;;
;;; TEXT & FORMATTING & LINKS
;;;

;;; <html-line-break> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-line-break> ()
  ((number-of-lines :initform 1 :accessor number-of-lines
		    :init-keyword :number-of-lines)))

(define-method write-html ((line-break <html-line-break>))
  (dotimes (i (number-of-lines line-break))
	   (format #t "<br>\n")))

(define-macro (html-line-break . args)
  `(make <html-line-break> ,@args))

;;; <html-paragraph-break> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-paragraph-break> ()
  ())

(define-method write-html ((paragraph <html-paragraph-break>))
  (format #t "<p>\n"))

(define-macro (html-paragraph-break . args)
  `(make <html-paragraph-break> ,@args))

;;; <html-paragraph> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; With STRING slot, it ignores CONTENTS slot.  This is for backward
;;; compatibility.

(define-class <html-paragraph> (<html-container>
                                <html-alignment-mixin>)
  ((string :accessor string :init-keyword :string)))

(define-method write-html ((paragraph <html-paragraph>))
  (format #t "<p")
  (write-html-tag-options paragraph)
  (format #t ">\n")
  (if (slot-bound? paragraph 'string)
      (format #t "~a\n" (safe-html-string (string paragraph)))
      (next-method))
  (format #t "</p>\n"))

(define-macro (html-paragraph . args)
  `(make <html-paragraph> ,@args))

;;; <html-center> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-center> (<html-container>)
  ())

(define-method write-html ((self <html-center>))
  (format #t "<center>\n")
  (next-method)
  (format #t "</center>\n"))

(define-macro (html-center . args)
  `(make <html-center> ,@args))

;;; <html-blockquote> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-blockquote> (<html-container>)
  ())

(define-method write-html ((self <html-blockquote>))
  (format #t "<blockquote>\n")
  (next-method)
  (format #t "</blockquote>\n"))

(define-macro (html-blockquote . args)
  `(make <html-blockquote> ,@args))

;;; <html-text> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-text> ()
  ((string :accessor string :initform "" :init-keyword :string)
   (bold? :initform #f :accessor bold? :init-keyword :bold?)
   (italic? :initform #f :accessor italic? :init-keyword :italic?)
   (underline? :initform #f :accessor underline? :init-keyword :underline?)
   (typewriter? :initform #f :accessor typewriter? :init-keyword :typewriter?)
   (preformatted? :initform #f :accessor preformatted? :init-keyword :preformatted?)
   (font-color :initform #f :accessor font-color :init-keyword :font-color)
   (font-size  :initform #f :accessor font-size :init-keyword :font-size)
   ))

(define-method write-html ((text <html-text>))
  (if (bold? text) (format #t "<B>"))
  (if (italic? text) (format #t "<I>"))
  (if (underline? text) (format #t "<U>"))
  (if (typewriter? text) (format #t "<TT>"))
  (if (preformatted? text) (format #t "<PRE>"))
  (if (or (font-color text) (font-size text))
      (begin (format #t "<FONT")
             (if (font-color text) (format #t " color=~s" (font-color text)))
             (if (font-size text) (format #t " size=~s" (font-size text)))
             (format #t ">")))
  (format #t "~a" (safe-html-string (string text)))
  (if (or (font-color text) (font-size text)) (format #t "</FONT>"))
  (if (preformatted? text) (format #t "</PRE>"))
  (if (bold? text) (format #t "</B>"))
  (if (italic? text) (format #t "</I>"))
  (if (underline? text) (format #t "</U>"))
  (if (typewriter? text) (format #t "</TT>")))

(define-macro (html-text . args)
  `(make <html-text> ,@args))

;;; <html-header> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-header> (<html-alignment-mixin>)
  ((string :accessor string :init-keyword :string)
   (header-size :initform 1 :accessor header-size :init-keyword :header-size)))

(define-method write-html ((header <html-header>))
  (format #t "<H~a" (header-size header))
  (write-html-tag-options header)
  (format #t ">~a</H~a>\n"
	  (safe-html-string (string header))
	  (header-size header)))

(define-macro (html-header . args)
  `(make <html-header> ,@args))

;;; <html-text-link> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; OBSOLETED.  Use <html-anchor> and <html-text>.

(define-class <html-text-link> (<html-text>)
  ((target :accessor target :init-keyword :target)))
  
(define-method write-html ((link <html-text-link>))
  (format #t "<A HREF=~s>~a</A>\n"
	  (target link) (safe-html-string (string link))))

(define-macro (html-text-link . args)
  `(make <html-text-link> ,@args))

;;;
;;; FORMS & INPUTS
;;;

;;; <html-form> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-form> (<html-container> <html-target-frame-mixin>)
  ((use-default-server? :initform #t :accessor use-default-server?
			:init-keyword :use-default-server?)
   (cgi-method :initform "get" :accessor cgi-method :init-keyword :cgi-method)
   (cgi-action :accessor cgi-action :init-keyword :cgi-action)
   ;; ((arg val) ...)
   (cgi-extra-args :initform #f :accessor cgi-extra-args
		   :init-keyword :cgi-extra-args)
   ))

(define-method write-html-tag-options ((form <html-form>))
  (format #t " method=~s action=~s" 
          (cgi-method form)
          (if (use-default-server? form)
	      (string-append *default-http-server* (cgi-action form))
	      (cgi-action form)))
  (next-method))

(define-method write-html ((form <html-form>))
  (format #t "<form")
  (write-html-tag-options form)
  (format #t ">\n")
  (if (cgi-extra-args form)
      (for-each (lambda (arg)
		  (format #t "<input type=\"hidden\" name=~s value=~s>\n"
			  (car arg) (cadr arg)))
		(cgi-extra-args form)))
  (next-method)
  (format #t "</form>\n"))

(define-macro (html-form . args)
  `(make <html-form> ,@args))

;;; <html-input> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-input> ()
  ((name :accessor name :init-keyword :name)
   (value :initform "" :accessor value :init-keyword :value)))

(define-macro (html-input . args)
  `(make <html-input> ,@args))

;;; <html-hidden> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-hidden> (<html-input>)
  ())

(define-method write-html ((input <html-hidden>))
  (format #t "<input type=\"hidden\" name=~s value=~s>\n"
	  (name input) (value input)))

(define-macro (html-hidden . args)
  `(make <html-hidden> ,@args))

;;; <html-submit-button> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-submit-button> (<html-input>)
  ())

(define-method write-html ((input <html-submit-button>))
  (format #t "<input type=\"submit\" name=~s value=~s>\n"
	  (name input) (value input)))

(define-macro (html-submit-button . args)
  `(make <html-submit-button> ,@args))

;;; <html-reset-button> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-reset-button> (<html-input>)
  ())

(define-method write-html ((input <html-reset-button>))
  (format #t "<input type=\"reset\" name=~s value=~s>\n"
	  (name input) (value input)))

(define-macro (html-reset-button . args)
  `(make <html-reset-button> ,@args))

;;; <html-text-field> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-text-field> (<html-input>)
  ((size :initform 40 :accessor size :init-keyword :size)))

(define-method write-html ((input <html-text-field>))
  (format #t "<input type=\"text\" name=~s value=~s size=~a>\n"
	  (name input) (value input) (size input)))

(define-macro (html-text-field . args)
  `(make <html-text-field> ,@args))

;;; <html-password> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-password> (<html-input>)
  ((size :initform 40 :accessor size :init-keyword :size)))

(define-method write-html ((input <html-password>))
  (format #t "<input type=\"password\" name=~s value=~s size= ~a>\n"
	  (name input) (value input) (size input)))

(define-macro (html-password . args)
  `(make <html-password> ,@args))

;;; <html-checkbox> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-checkbox> (<html-input>)
  ((checked? :initform #f :accessor checked? :init-keyword :checked?)))

(define-method write-html ((input <html-checkbox>))
  (format #t "<input type=\"checkbox\" name=~s value=~s~a>\n"
	  (name input) (value input) (if (checked? input) " checked" "")))

(define-macro (html-checkbox . args)
  `(make <html-checkbox> ,@args))

;;; <html-radio> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-radio> (<html-input>)
  ((checked? :initform #f :accessor checked? :init-keyword :checked?)))

(define-method write-html ((input <html-radio>))
  (format #t "<input type=\"radio\" name=~s value=~s~a>\n"
	  (name input) (value input) (if (checked? input) " checked" "")))

(define-macro (html-radio . args)
  `(make <html-radio> ,@args))

;;; <html-menu> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-menu> (<html-input>)
  ((options :accessor options :init-keyword :options)))

(define-method write-html-options ((menu <html-menu>))
  (map (lambda (item)
         (cond ((string? item)
                (format #t "<option value=~s>~a\n" item item))
               ((list? item)
                (format #t "<option value=~s~a>~a\n"
                        (car item)
                        (if (and (= (length item) 3)
                                 (eq? (caddr item) 'selected))
                            " selected" "")
                        (cadr item)))))
       (options menu)))

(define-method write-html ((menu <html-menu>))
  (format #t "<select name=~s>\n" (name menu))
  (write-html-options menu)
  (format #t "</select>\n"))

(define-macro (html-menu . args)
  `(make <html-menu> ,@args))

;;; <html-active-menu> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-active-menu> (<html-menu>)
  ((script :accessor script :init-keyword :script)))

(define-method write-html ((menu <html-active-menu>))
  (write-html (script menu))
  (format #t "<select name=~s onchange=\"~a(this.form, this)\">\n"
	  (name menu) (script-name menu))
  (write-html-options menu)
  (format #t "</select>\n"))

(define-macro (html-active-menu . args)
  `(make <html-active-menu> ,@args))

;;; <html-cgi-menu> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define *html-classes:cgi-menu-dispatch-script-format-string*
  "function ~a(form, selection) {
     for (i=0; i<selection.options.length; i++) {
       if (selection.options[i].selected == true) {
         val = selection.options[i].value;
       }
     }
     locCgi = \"~a?value=\";
     locCal = locCgi + val;
     location.href = locCal;
   }")

(define-class <html-cgi-menu> (<html-active-menu>)
  ((cgi-string :accessor cgi-string :init-keyword :cgi-string)))

(define-method write-html ((menu <html-active-menu>))
  (let ((tmp-name (symbol->string (gensym "javascript_"))))
    (set! (script menu)
	  (make <html-java-script>
		:script-name tmp-name
		:script-body (format #f *html-classes:cgi-menu-dispatch-script-format-string*
				     tmp-name (cgi-string menu))))
    (write-html (script menu))
    (format #t "<select name=~s onchange=\"~a(this.form, this)\">\n"
	    (name menu) tmp-name)
    (write-html-options menu)
    (format #t "</select>\n")))

(define-macro (html-cgi-menu . args)
  `(make <html-cgi-menu> ,@args))

;;; <html-titled-menu> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-titled-menu> (<html-menu>)
  ((menu-title :accessor menu-title :init-keyword :menu-title)))

(define-method write-html ((menu <html-titled-menu>))
  (format #t "~a &nbsp;&nbsp;&nbsp;" (menu-title menu))
  (next-method))

(define-macro (html-titled-menu . args)
  `(make <html-titled-menu> ,@args))

;;; <html-scrolled-list> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-scrolled-list> (<html-menu>)
  ((size :initform 5 :accessor size :init-keyword :size)
   (multiple? :initform #f :accessor multiple? :init-keyword :multiple?)))

(define-method write-html ((scrolled-list <html-scrolled-list>))
  (format #t "<select name=~s size=~a ~a>\n"
	  (name scrolled-list) (size scrolled-list)
	  (if (multiple? scrolled-list) "multiple" ""))
  (write-html-options scrolled-list)
  (format #t "</select>\n"))

(define-macro (html-scrolled-list . args)
  `(make <html-scrolled-list> ,@args))

;;; <html-text-area> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-text-area> (<html-input>)
  ((rows :initform 5 :accessor rows :init-keyword :rows)
   (columns :initform 40 :accessor columns :init-keyword :columns)
   (wrap :initform #f :accessor wrap :init-keyword :wrap)))
  
(define-method write-html ((text-area <html-text-area>))
  (format #t "<textarea name=\"~a\" rows=~a cols=~a~a>~a</textarea>\n"
	  (name text-area)
          (rows text-area)
          (columns text-area)
          (if (wrap text-area) (format #f " wrap=~s" (wrap text-area)) "")
          (or (value text-area) "")))

(define-macro (html-text-area . args)
  `(make <html-text-area> ,@args))

;;;
;;; TABLES
;;;

;;; <html-table> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; NOTE: field BORDER? is left for backward compatibility.

(define-class <html-table> (<html-container> <html-border-mixin>
                                             <html-alignment-mixin>
                                             <html-size-mixin>)
  ((caption :initform #f :accessor caption :init-keyword :caption)
   (border? :initform #t :accessor border? :init-keyword :border?)
   (cell-padding :initform 2 :accessor cell-padding :init-keyword :cell-padding)
   (cell-spacing :initform 0 :accessor cell-spacing :init-keyword :cell-spacing)))

(define-method write-html-tag-options ((table <html-table>))
  (if (cell-padding table) (format #t " cellpadding=~s" (cell-padding table)))
  (if (cell-spacing table) (format #t " cellspacing=~s" (cell-spacing table)))
  (next-method))

(define-method write-html ((table <html-table>))
  (if (border? table) (set! (border table) 1))
  (format #t "<table")
  (write-html-tag-options table)
  (format #t ">\n")
  (if (caption table)
      (format #t "<caption><B>~a</B></caption>\n" (caption table)))
  (next-method)
  (format #t "</table>\n"))

(define-macro (html-table . args)
  `(make <html-table> ,@args))

;;; data = '((obj obj ...) (obj obj ...) ...)

(define (make-layout-table caption data)
  (make <html-table>
	:border? #f
	:caption caption
	:contents
	(map (lambda (row)
	       (make <html-table-row>
		     :contents (map (lambda (cell)
				      (make <html-table-cell>
					    :alignment (if (eq? cell (car row))
							   "right"
							   "left")
					    :contents (list cell)))
				    row)))
	       data)))

;;; <html-table-row> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-table-row> (<html-container> <html-alignment-mixin>)
  ())
  
(define-method write-html ((table-row <html-table-row>))
  (format #t "<tr")
  (write-html-tag-options table-row)
  (format #t ">")
  (next-method)
  (format #t "</tr>\n"))

(define-macro (html-table-row . args)
  `(make <html-table-row> ,@args))

;;; <html-table-header> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-table-header> (<html-container>
                                   <html-alignment-mixin>
                                   <html-bgcolor-mixin>
                                   <html-size-mixin>
                                   <html-table-element-mixin>)
  ())
  
(define-method write-html ((table-header <html-table-header>))
  (format #t "<th")
  (write-html-tag-options table-header)
  (format #t ">\n")
  (next-method)
  (format #t "</th>\n"))

(define-macro (html-table-header . args)
  `(make <html-table-header> ,@args))

;;; <html-table-cell> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-table-cell> (<html-container>
                                   <html-alignment-mixin>
                                   <html-bgcolor-mixin>
                                   <html-size-mixin>
                                   <html-table-element-mixin>)
  ())
  
(define-method write-html ((table-cell <html-table-cell>))
  (format #t "<td")
  (write-html-tag-options table-cell)
  (format #t ">\n")
  (next-method)
  (format #t "</td>\n"))

(define-macro (html-table-cell . args)
  `(make <html-table-cell> ,@args))

;;;
;;; FRAMES
;;;

;;; <html-frameset> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; NOTE: FRAMESET tag definition in HTML4.0 specification doesn't
;; have FRAMEBORDER and BORDER tag.  However, Netscape accepts them
;; and lots of websites are actually using them.

(define-class <html-frameset> (<html-container> <html-border-mixin>)
  ((rows     :accessor rows :init-keyword :rows :initform #f)
   (cols     :accessor cols :init-keyword :cols :initform #f)
   (onload   :accessor onload :init-keyword :onload :initform #f)
   (onunload :accessor onunload :init-keyword :onunload :initform #f)
   (frame-border :accessor frame-border :init-keyword :frame-border
                 :initform #f)
   ))

(define-method write-html-tag-options ((self <html-frameset>))
  (if (rows self) (format #t " rows=~s" (rows self)))
  (if (cols self) (format #t " cols=~s" (cols self)))
  (if (onload self) (format #t " onload=~s" (onload self)))
  (if (onunload self) (format #t " onunload=~s" (onunload self)))
  (if (frame-border self) (format #t " frameborder=~s" (frame-border self)))
  (next-method)
  )

(define-method write-html ((self <html-frameset>))
  (format #t "<frameset")
  (write-html-tag-options self)
  (format #t ">\n")
  (next-method)
  (format #t "</frameset>\n"))

(define-macro (html-frameset . args)
  `(make <html-frameset> ,@args))

;;; <html-frame> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-frame> ()
  ((source    :accessor source :init-keyword :source :initform "")
   (name      :accessor name :init-keyword :name :initform #f)
   (frame-border :accessor frame-border :init-keyword :frame-border
                 :initform #f)
   (margin-width :accessor margin-width :init-keyword :margin-width
                 :initform #f)
   (margin-height :accessor margin-height :init-keyword :margin-height
                 :initform #f)
   (scrolling :accessor scrolling :init-keyword :scrolling :initform #f)
   (no-resize? :accessor no-resize? :init-keyword :no-resize :initform #f)))

(define-method write-html-tag-options ((self <html-frame>))
  (if (source self) (format #t " src=~s" (source self)))
  (if (name self)   (format #t " name=~s" (name self)))
  (if (frame-border self)
      (format #t " frame-boder=~s" (frame-border self)))
  (if (margin-width self)
      (format #t " marginwidth=~s" (margin-width self)))
  (if (margin-height self)
      (format #t " marginheight=~s" (margin-height self)))
  (if (scrolling self)
      (format #t " scrolling=~s" (scrolling self)))
  (if (no-resize? self)
      (format #t " noresize")))

(define-method write-html ((self <html-frame>))
  (format #t "<frame")
  (write-html-tag-options self)
  (format #t ">\n"))

(define-macro (html-frame . args)
  `(make <html-frame> ,@args))

;;; <html-noframes> ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <html-noframes> (<html-container>)
  ())

(define-method write-html ((self <html-noframes>))
  (format #t "<noframes")
  (write-html-tag-options self)
  (format #t ">\n")
  (next-method)
  (format #t "</noframes>\n"))

(define-macro (html-noframes . args)
  `(make <html-noframes ,@args))

;;; test ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (test-html)
  (let ((page
	 (make <html-page>
	       :page-title "Schmtl Test Page"
	       :background-color "white"
	       :author-email "kaveh@squareusa.com"
	       :author-name "Kaveh Kardan"
	       :contents
	       (list
		(make <html-paragraph>
		      :string "This is a test of \"Schmtl\"<br> the scheme html generation system.")
		(make <html-form>
		      :cgi-method "get"
		      :cgi-action "/gray/db/cgi/foo.cgi"
		      :contents
		      (list
		       (make <html-table>
			     :caption "This is a table, dudes."
			     :border? #t
			     :contents
			     (list
			      (make <html-table-row>
				    :contents (list
					       (make <html-table-header>
						     :contents
						     (list
						      (make <html-text>
							    :string "")))
					       (make <html-table-header>
						     :contents
						     (list
						      (make <html-text>
							    :string "col 1")))
					       (make <html-table-header>
						     :contents
						     (list
						      (make <html-text>
							    :string "col 2")))
					       (make <html-table-header>
						     :contents
						     (list
						      (make <html-text>
							    :string "col 3")))))
			      (make <html-table-row>
				    :contents (list
					       (make <html-table-header>
						     :contents
						     (list
						      (make <html-text>
							    :string "row #1")))
					       (make <html-table-cell>
						     :contents
						     (list
						      (make <html-text>
							    :string "1")))
					       (make <html-table-cell>
						     :contents
						     (list
						      (make <html-text-link>
							    :string "2"
							    :target "/www/texture/create/isaka/WEB/central/")))
					       (make <html-table-cell>
						     :contents
						     (list
						      (make <html-text>
							    :string "3")))))
			      (make <html-table-row>
				    :contents (list
					       (make <html-table-header>
						     :contents
						     (list
						      (make <html-text>
							    :string "numero 2")))
					       (make <html-table-cell>
						     :contents
						     (list
						      (make <html-text>
							    :string "a")))
					       (make <html-table-cell>
						     :contents
						     (list
						      (make <html-text>
							    :string "b")))
					       (make <html-table-cell>
						     :contents
						     (list
						      (make <html-text>
							    :string "c")))))
			      (make <html-table-row>
				    :contents (list
					       (make <html-table-header>
						     :contents
						     (list
						      (make <html-text>
							    :string "The Third Row")))
					       (make <html-table-cell>
						     :contents
						     (list
						      (make <html-text>
							    :string "Alpha")))
					       (make <html-table-cell>
						     :contents
						     (list
						      (make <html-menu>
							    :name "test_menu"
							    :options '("Choice 1" "Choice 2" "and 3"))))
					       (make <html-table-cell>
						     :contents
						     (list
						      (make <html-text>
							    :string "Charlie")))))))
		       (make <html-titled-menu>
			     :menu-title "My Menu"
			     :name "test_menu_2"
			     :options '("Choice 1" "Choice 2" "and 3"))
		       (make <html-paragraph-break>)
		       (make <html-text-field>
			     :name "text-field"
			     :value "text field"
			     :size 10)
		       (make <html-password>
			     :name "password")
		       (make <html-paragraph>
			     :string "Man, it's good to be back writing Scheme!")
		       (make <html-text-area>
			     :name "comments"
			     :rows 1 :columns 40)
		       (make <html-paragraph-break>)
		       (make <html-submit-button>
			     :name "submit"
			     :value "OK")
		       (make <html-reset-button>
			     :name "reset"
			     :value "Clear")
		       ))))))
    (with-output-to-file "test.html"
      (lambda () (write-html page)))))


